from pathlib import Path
import re
from shutil import rmtree
import string
import random
from werkzeug.utils import secure_filename
from git import Repo, InvalidGitRepositoryError, GitCommandError
from flask import current_app as app

from biscd import config
from .yaml_serializable import YamlSerializable
from .utils import to_bool

class Project(YamlSerializable):

    #@property
    @classmethod
    def _storage_file(cls):
        return Path(config.config_dir()) / 'projects.yaml'

    #@property
    @classmethod
    def _yaml_object_name(cls):
        return 'projects'

    def __init__(self, name=None, branch='master', git_repo = None):
        super().__init__()
        self.name = name
        self.branch = branch
        self.relative_production_path = None
        self.git_repo = git_repo
        self.access = {'public': False, 'owners': []}

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Project):
            if self.name == other.name:
                return True
        return False

    def __hash__(self):
        return hash(self.name)

    @property
    def absolute_production_path(self):
        if not hasattr(self, 'relative_production_path') or not self.relative_production_path:
            return None
        production_path = Path(config['production_path'].get())
        absolute_path = (production_path / self.relative_production_path).resolve()
        try:
            if str(absolute_path.relative_to(production_path)) != self.relative_production_path:
                raise ValueError
        except ValueError as v_e:
            raise ValueError(f"{self.name}'s relative_production_path is not within the " \
                    f"application's production path! {self.relative_production_path}") from v_e
        return absolute_path

    def update(self):
        """Update project's local repoistory"""
        # pylint: disable=no-member
        self._make_sure_production_path_exists()
        path = self.absolute_production_path
        app.logger.info("Updating project '%s' in %s", self.name, path)
        try:
            repo = Repo(path)
            app.logger.info('%s was already under git control', self.name)
        except InvalidGitRepositoryError:
            repo = self._clone_repo(path)
            # If repo is not type Repo it must be an error and thus returned directly
            if not isinstance(repo, Repo):
                return repo
        remotes = repo.remotes
        if not any(remotes):
            app.logger.error("%s's repo %s doesn't have any remotes.", self.name, path)
            return ["Repo doesn't have any remotes. please fix manually", 'error']
        response = remotes[0].pull()
        for item in response:
            #TODO: is this nessecary?
            app.logger.debug(item)

        response = self._checkout_branch(repo, self.branch)
        return response

    def _clone_repo(self, path):
        app.logger.info("Cloning repo from %s", self.git_repo)
        try:
            repo = Repo.clone_from(self.git_repo, path)
        except GitCommandError as gce:
            app.logger.error("Project %s's git repo: %s is invalid: %s",
                             self.name, self.git_repo, gce.stderr.strip('\n '))
            return ["Project does not have a valid git repository url", "error"]
        app.logger.info('Cloning repo (%s) succeeded!', self.git_repo)
        return repo

    def _checkout_branch(self, repo, branch):
        try:
            repo.git.checkout(branch)
        except AttributeError:
            return [f"Branch is not set for {self.name}", 'error']
        except GitCommandError:
            return [
                f"Repo '{self.git_repo}' does not have an existing branch called {branch}",
                'error']
        return None

    def delete(self):
        if self.production_path_exists():
            result = self.delete_files()
            if result and result[1] in ['warning', 'error']:
                return result
        super().delete()
        return [f"{self.name} has been deleted", 'success']

    def delete_files(self):
        if not hasattr(self, 'relative_production_path') or not self.relative_production_path:
            return ['Project does not have a production path', 'warning']
        if self.production_path_exists():
            path = self.absolute_production_path
            app.logger.info(f"Deleting {path}")
            rmtree(path)
        path = Path(self.relative_production_path).parent
        while len(path.parts) > 0 and self._delete_dir_if_empty(path):
            path = path.parent
        self.relative_production_path = None
        self.save()

    @classmethod
    def _delete_dir_if_empty(cls, relative_path):
        production_path = config['production_path'].get()
        absolute_path = (production_path / relative_path).resolve()
        if absolute_path.is_dir() and not any(absolute_path.iterdir()):
            if absolute_path.relative_to(production_path) == relative_path:
                absolute_path.rmdir()
                app.logger.debug('Deleted empty dir %s', absolute_path)
                return True
            raise ValueError("Trying to delete a directory outdize of production path! : "
                              + str(absolute_path))
        return False

    def production_path_exists(self):
        """Return True if project's the production directory path exists"""
        if hasattr(self, 'relative_production_path') and self.relative_production_path:
            absolute_path = self.absolute_production_path
            return absolute_path.is_dir()
        return False

    def _make_sure_production_path_exists(self):
        if hasattr(self, 'relative_production_path') and self.relative_production_path:
            absolute_path = self.absolute_production_path
            if not absolute_path.is_dir():
                absolute_path.mkdir(parents=True)
        else:
            prod_path_dir = Path(config['production_path'].get())
            def generate_absolute_path():
                rand_num = ''.join(random.choices(string.digits, k=6))
                return prod_path_dir / (secure_filename(self.name) + rand_num)

            absolute_path = generate_absolute_path()
            # Make sure path is unique
            while absolute_path.is_dir():
                absolute_path = generate_absolute_path()

            absolute_path.mkdir(parents=True)
            self.relative_production_path = str(absolute_path.relative_to(prod_path_dir))
            self.save()

    def user_access(self, user):
        # pylint: disable=no-member
        if not hasattr(self, 'access') or self.access is None:
            return None
        if hasattr(user,('name')) and user.name in self.access.get('owners', []):
            return 'Owner'
        public = to_bool(self.access.get('public', False))
        if public:
            return 'Public'
        return None
