from abc import ABCMeta, abstractmethod
from flask import abort
from git import exc
import yaml
import inspect

from .recursive_property import RecursiveProperty

class MyMeta(metaclass=ABCMeta):
    required_attributes = []

    def __call__(self, *args, **kwargs):
        # pylint: disable=no-member
        obj = super(MyMeta, self).__call__(*args, **kwargs)
        for attr_name in obj.required_attributes:
            if not getattr(obj, attr_name):
                raise ValueError('required attribute (%s) not set' % attr_name)
        return obj

class YamlSerializable(RecursiveProperty):
    __metaclass__ = MyMeta
    required_attributes = ['name']

    def __init__(self):
        super().__init__('name')

    @classmethod
    #@property
    @abstractmethod
    def _storage_file(cls):
        pass

    @classmethod
    #@property
    @abstractmethod
    def _yaml_object_name(cls):
        pass

    def save(self, overwrite=True):
        # pylint: disable=no-member
        if self.name is None:
            raise TypeError("Name cannot be None")
        ymlsls = self._get_all_from_file()
        ymlsl = next((ymlsl for ymlsl in ymlsls if [*ymlsl][0] == self.name), None)
        if ymlsl:
            if overwrite:
                ymlsl[self.name] = self.config_dict(values_only=True)
            else:
                raise ValueError(
                    f"A {type(self).__name__} with name {self.name} already exists!")
        else:
            ymlsls.append(self.config_dict())
            print(ymlsls)
        self._save_all_to_file(ymlsls)

    def delete(self):
        # pylint: disable=no-member
        if self.name is None:
            raise TypeError("Name cannot be None")
        ymlsls = self._get_all_from_file()
        try:
            ymlsl = next((ymlsl for ymlsl in ymlsls if [*ymlsl][0] == self.name))
            ymlsls.remove(ymlsl)
        #ymlsl = next((ymlsl for ymlsl in ymlsls if [*ymlsl][0] == self.name), None)
        except KeyError as key_error:
            raise ValueError(
                f"{type(self).__name__} {self.name} does not exist") from key_error
        self._save_all_to_file(ymlsls)

    @classmethod
    def first_or_404(cls, **kwargs):
        ymlsl = cls.first(**kwargs)
        if ymlsl is None:
            abort(404)
        return ymlsl

    @classmethod
    def first(cls, **kwargs):
        ymlsls = cls.get(**kwargs)
        if not any(ymlsls):
            return None
        return ymlsls[0]

    @classmethod
    def get(cls, **kwargs):
        """Returns any matching instances

        Filters all saved instances by specified properties.
        All remaining items are returned as a list.
        """

        if not any(kwargs):
            return []
        ymlsl_dicts = cls._get_all_from_file()
        for key, value in kwargs.items():
            ymlsl_dicts = cls.filter_dicts(ymlsl_dicts, key, value)

        ymlsls = []
        for ymlsl_dict in ymlsl_dicts:
            ymlsls.append(cls._from_dict(ymlsl_dict, 'name'))
        return ymlsls

    @staticmethod
    def filter_dicts(ymlsl_dicts, key, value):
        # 'name' has to be evaluated separately; 'name' is the key of the entire object
        if key == 'name':
            return [ymlsl_dict for ymlsl_dict in ymlsl_dicts
                            if [*ymlsl_dict][0] == value]

        # For other keys, filter out any item that does not contain the key, or that not match
        # the key's value

        # If key is recursive
        if '__' in key:
            return [ymlsl_dict for ymlsl_dict in ymlsl_dicts
                    if YamlSerializable.has_recursive_property(
                        list(ymlsl_dict.items())[0][1], key, value)]

        # If key is normal key
        return [ymlsl_dict for ymlsl_dict in ymlsl_dicts
                if next(iter(ymlsl_dict.values())).get(key, None) == value]

    @staticmethod
    def has_recursive_property(dictionary, key, value):
        elements = key.split('__')
        for element in elements:
            # If last element: check if last element equals value
            if element is elements[-1]:
                if isinstance(dictionary, list):
                    item = [item for item in dictionary if list(item.keys())[0] == element][0]
                    if item[element] == value:
                        return True
                    return False
                element_in_dict = dictionary.get(element, None)
                if isinstance(element_in_dict, list):
                    if value in element_in_dict:
                        return True
                    return False
                if element_in_dict == value:
                    return True
                return False

            # Else check if dictionary has element. If so, dictionary becomes element
            if isinstance(dictionary, list):
                item = [item for item in element_in_dict if list(item.keys())[0] == element][0]
                if item[element] is not None:
                    dictionary = item[element]
                    continue
            element_in_dict = dictionary.get(element, None)
            if isinstance(dictionary, list):
                if element in [list(item.keys())[0] for item in element_in_dict]:
                    continue
                return False
            if element_in_dict is None:
                return False
            dictionary = element_in_dict
        return False

    @classmethod
    def list_names(cls):
        ymlsls = cls._get_all_from_file()
        ymlserializables_list = []
        for ymlsl in ymlsls:
            name = [*ymlsl][0]
            ymlserializables_list.append(name)
        return ymlserializables_list


    @classmethod
    def _get_all_from_file(cls):
        with open(cls._storage_file()) as file:
            ymlsls = yaml.load(file, yaml.FullLoader).get(cls._yaml_object_name())
        return ymlsls

    @classmethod
    def _save_all_to_file(cls, ymlsls):
        ymlsls_object = {cls._yaml_object_name() : ymlsls}
        with open(cls._storage_file(), 'w') as file:
            yaml.dump(ymlsls_object, file)
