from flask import render_template, flash, abort, redirect, request, url_for
from flask_login import current_user, login_user, logout_user, login_required
from werkzeug.urls import url_parse
from biscd import app
from .utils import flash_result
from .models import Project, User
from .froms import ProjectForm, LoginForm, RegistrationForm

@app.route('/', methods=['GET', 'POST'])
@app.route('/index', methods=['GET', 'POST'])
def index():
    if current_user.is_authenticated:
        projects = current_user.projects
    else:
        projects = None
    return render_template('index.html', projects=projects)

@app.route('/login', methods=['GET', 'POST'])
def login():
    if current_user.is_authenticated:
        return redirect(url_for('index'))
    form = LoginForm()
    if form.validate_on_submit():
        user = User.first(name=form.username.data)
        if user is None or not user.check_password(form.password.data):
            flash('Invalid username or password', 'warning')
            return redirect(url_for('login'))
        login_user(user, remember=form.remember_me.data)
        next_page = request.args.get('next')
        if not next_page or url_parse(next_page).netloc != '':
            next_page = url_for('index')
        return redirect(next_page)
    return render_template('login.html', title='Sign In', form=form)

@app.route('/logout')
@login_required
def logout():
    logout_user()
    return redirect(url_for('index'))

@app.route('/register', methods=['GET', 'POST'])
def register():
    if current_user.is_authenticated:
        return redirect(url_for('index'))
    form = RegistrationForm()
    if form.validate_on_submit():
        user = User(name=form.username.data, email=form.email.data)
        user.set_password(form.password.data)
        user.save()
        flash('Congratulations, you are now a registered user!', 'success')
        return redirect(url_for('login'))
    return render_template('register.html', title='Register', form=form)

@app.route('/project/public', methods=['GET'])
def public_projects():
    projects = Project.get(access__public = True)
    return render_template('projects.html', title='Public Projects', projects=projects)

@app.route('/project/add', methods=['GET', 'POST'])
@login_required
def project_add():
    project = Project(name='')
    form = ProjectForm()
    if form.validate_on_submit():
        project = Project(name=form.projectname.data, git_repo=form.git_repo.data)
        project.access['public'] = form.public.data
        if project.access['owners'] == []:
            project.access['owners'] = [current_user.name]
        project.save()
        flash('Your project is created!', 'success')
        return redirect(url_for('project_dashboard', project_name=project.name))
    return render_template('project_settings.html', title='Add Project', form=form, project=project)

@app.route('/project/<project_name>/settings', methods=['GET', 'POST'])
@login_required
def project_change_settings(project_name):
    project = Project.first_or_404(name=project_name)
    if project.user_access(current_user) != 'Owner':
        abort(401)
    form = ProjectForm()
    form.old_name = project.name
    if request.method == 'GET':
        # pylint: disable=no-member
        form.process()
        form.public.data = project.access.get('public', False)
    if form.validate_on_submit():
        project = Project(name=form.projectname.data, git_repo=form.git_repo.data)
        project.access['public'] = form.public.data
        if project.access['owners'] == []:
            project.access['owners'] = [current_user.name]
        project.save()
        flash(f"{project.name} was updated!", 'success')
        return redirect(url_for('project_dashboard', project_name=project.name))

    return render_template('project_settings.html', title=f'{project.name}: Settings', form=form,
                           project=project)

@app.route('/project/<project_name>', methods=['GET'])
def project_dashboard(project_name):
    project = Project.first_or_404(name=project_name)
    access = project.user_access(current_user)
    if access is None:
        abort(404)
    return render_template('project.html', title=f'{project.name}', project=project, owner=(access == 'Owner'))

@app.route('/project/<project_name>/update', methods=['GET'])
@login_required
def project_update(project_name):
    project = Project.first_or_404(name=project_name)
    if project.user_access(current_user) != 'Owner':
        abort(401)
    result = project.update()
    flash_result(result)
    return redirect(url_for('project_dashboard', project_name=project.name))

@app.route('/project/<project_name>/delete_files', methods=['GET'])
@login_required
def project_delete_files(project_name):
    project = Project.first_or_404(name=project_name)
    if project.user_access(current_user) != 'Owner':
        abort(401)
    result = project.delete_files()
    flash_result(result)
    return redirect(url_for('project_dashboard', project_name=project.name))

@app.route('/project/<project_name>/delete', methods=['GET'])
@login_required
def project_delete(project_name):
    project = Project.first_or_404(name=project_name)
    if project.user_access(current_user) != 'Owner':
        abort(401)
    result = project.delete()
    flash_result(result)
    return redirect(url_for('index'))
